@@ -26,8 +26,6 @@
 
 #define G_SETTINGS_ENABLE_BACKEND
 
-#include <dbus/dbus-glib-bindings.h>
-#include <dbus/dbus-glib.h>
 #include <gio/gio.h>
 #include <gio/gsettingsbackend.h>
 #include <glib/gi18n.h>
@@ -38,7 +36,6 @@
 #include "yelp-view.h"
 
 #include "yelp-application.h"
-#include "yelp-dbus.h"
 #include "yelp-window.h"
 
 #define DEFAULT_URI "ghelp:user-guide"
@@ -47,6 +44,7 @@ static gboolean editor_mode = FALSE;
 
 enum {
     BOOKMARKS_CHANGED,
+    READ_LATER_CHANGED,
     LAST_SIGNAL
 };
 static gint signals[LAST_SIGNAL] = { 0 };
@@ -63,11 +61,32 @@ struct _YelpApplicationLoad {
     gboolean new;
 };
 
+static const gchar introspection_xml[] =
+    "<node name='/org/gnome/Yelp'>"
+    "  <interface name='org.gnome.Yelp'>"
+    "    <annotation name='org.freedesktop.DBus.GLib.CSymbol' value='yelp_application'/>"
+    "    <method name='LoadUri'>"
+    "      <arg type='s' name='Uri' direction='in'/>"
+    "      <arg type='u' name='Timestamp' direction='in'/>"
+    "    </method>"
+    "  </interface>"
+    "</node>";
+GDBusNodeInfo *introspection_data;
+
 static void          yelp_application_init             (YelpApplication       *app);
 static void          yelp_application_class_init       (YelpApplicationClass  *klass);
 static void          yelp_application_dispose          (GObject               *object);
 static void          yelp_application_finalize         (GObject               *object);
 
+static void          yelp_application_method           (GDBusConnection       *connection,
+                                                        const gchar           *sender,
+                                                        const gchar           *object_path,
+                                                        const gchar           *interface_name,
+                                                        const gchar           *method_name,
+                                                        GVariant              *parameters,
+                                                        GDBusMethodInvocation *invocation,
+                                                        YelpApplication       *app);
+
 static void          application_setup                 (YelpApplication       *app);
 static void          application_uri_resolved          (YelpUri               *uri,
                                                         YelpApplicationLoad   *data);
@@ -87,16 +106,20 @@ static gboolean      window_resized                    (YelpWindow            *w
 G_DEFINE_TYPE (YelpApplication, yelp_application, G_TYPE_OBJECT);
 #define GET_PRIV(object)(G_TYPE_INSTANCE_GET_PRIVATE ((object), YELP_TYPE_APPLICATION, YelpApplicationPrivate))
 
+GDBusInterfaceVTable yelp_dbus_vtable = {
+    (GDBusInterfaceMethodCallFunc) yelp_application_method,
+    NULL, NULL };
+
 typedef struct _YelpApplicationPrivate YelpApplicationPrivate;
 struct _YelpApplicationPrivate {
-    DBusGConnection *connection;
+    GDBusConnection *connection;
 
     GSList *windows;
     GHashTable *windows_by_document;
 
     GtkActionGroup *action_group;
 
-    gchar *gsettings_context; /* static, do not free */
+    GSettingsBackend *backend;
     GSettings *gsettings;
     GHashTable *docsettings;
 };
@@ -139,8 +162,13 @@ yelp_application_class_init (YelpApplicationClass *klass)
                       g_cclosure_marshal_VOID__STRING,
                       G_TYPE_NONE, 1, G_TYPE_STRING);
 
-    dbus_g_object_type_install_info (YELP_TYPE_APPLICATION,
-                                     &dbus_glib_yelp_object_info);
+    signals[READ_LATER_CHANGED] =
+        g_signal_new ("read-later-changed",
+                      G_TYPE_FROM_CLASS (klass),
+                      G_SIGNAL_RUN_LAST,
+                      0, NULL, NULL,
+                      g_cclosure_marshal_VOID__STRING,
+                      G_TYPE_NONE, 1, G_TYPE_STRING);
 
     g_type_class_add_private (klass, sizeof (YelpApplicationPrivate));
 }
@@ -196,10 +224,9 @@ application_setup (YelpApplication *app)
         gchar *keyfile = g_build_filename (g_get_user_config_dir (),
                                            "yelp", "yelp.cfg",
                                            NULL);
-        g_settings_backend_setup_keyfile ("yelp", keyfile);
-        priv->gsettings_context = "yelp";
-        priv->gsettings = g_settings_new_with_context ("org.gnome.yelp",
-                                                       priv->gsettings_context);
+        priv->backend = g_keyfile_settings_backend_new (keyfile);
+        priv->gsettings = g_settings_new_with_backend ("org.gnome.yelp",
+                                                       priv->backend);
         g_free (keyfile);
     }
     else {
@@ -289,8 +316,8 @@ yelp_application_run (YelpApplication  *app,
 {
     GOptionContext *context;
     GError *error = NULL;
-    DBusGProxy *proxy;
-    guint request;
+    GVariant *ret;
+    guint32 request;
     YelpApplicationPrivate *priv = GET_PRIV (app);
     gchar *uri;
 
@@ -301,38 +328,38 @@ yelp_application_run (YelpApplication  *app,
     g_option_context_add_main_entries (context, entries, GETTEXT_PACKAGE);
     g_option_context_parse (context, &argc, &argv, NULL);
 
-    priv->connection = dbus_g_bus_get (DBUS_BUS_SESSION, &error);
+    priv->connection = g_bus_get_sync (G_BUS_TYPE_SESSION, NULL, &error);
     if (priv->connection == NULL) {
         g_warning ("Unable to connect to dbus: %s", error->message);
         g_error_free (error);
         return 1;
     }
 
-    /* FIXME: canonicalize relative URI */
     if (argc > 1)
         uri = argv[1];
     else
         uri = DEFAULT_URI;
 
-    proxy = dbus_g_proxy_new_for_name (priv->connection,
-                                       DBUS_SERVICE_DBUS,
-                                       DBUS_PATH_DBUS,
-                                       DBUS_INTERFACE_DBUS);
-
-    if (!org_freedesktop_DBus_request_name (proxy,
-                                            "org.gnome.Yelp",
-                                            0, &request,
-                                            &error)) {
+    ret = g_dbus_connection_call_sync (priv->connection,
+                                       "org.freedesktop.DBus",
+                                       "/org/freedesktop/DBus",
+                                       "org.freedesktop.DBus",
+                                       "RequestName",
+                                       g_variant_new ("(su)", "org.gnome.Yelp", 0),
+                                       NULL,
+                                       G_DBUS_CALL_FLAGS_NONE,
+                                       -1, NULL, &error);
+    if (error) {
         g_warning ("Unable to register service: %s", error->message);
         g_error_free (error);
-        g_object_unref (proxy);
+        g_variant_unref (ret);
         return 1;
     }
 
-    g_object_unref (proxy);
+    g_variant_get (ret, "(u)", &request);
+    g_variant_unref (ret);
 
-    if (request == DBUS_REQUEST_NAME_REPLY_EXISTS ||
-        request == DBUS_REQUEST_NAME_REPLY_IN_QUEUE) {
+    if (request == 2 || request == 3) { /* IN_QUEUE | EXISTS */
         gchar *newuri;
 
         if (uri && (strchr (uri, ':') || (uri[0] == '/')))
@@ -346,28 +373,32 @@ yelp_application_run (YelpApplication  *app,
             g_free (cur);
         }
 
-        proxy = dbus_g_proxy_new_for_name (priv->connection,
+        ret = g_dbus_connection_call_sync (priv->connection,
                                            "org.gnome.Yelp",
                                            "/org/gnome/Yelp",
-                                           "org.gnome.Yelp");
-        if (!dbus_g_proxy_call (proxy, "LoadUri", &error,
-                                G_TYPE_STRING, newuri,
-                                G_TYPE_UINT,
-                                gtk_get_current_event_time (),
-                                G_TYPE_INVALID, G_TYPE_INVALID)) {
+                                           "org.gnome.Yelp",
+                                           "LoadUri",
+                                           g_variant_new ("(su)", newuri, gtk_get_current_event_time ()),
+                                           NULL,
+                                           G_DBUS_CALL_FLAGS_NONE,
+                                           -1, NULL, &error);
+        if (error) {
             g_warning ("Unable to notify existing process: %s\n", error->message);
             g_error_free (error);
         }
 
         if (newuri != uri)
             g_free (newuri);
-        g_object_unref (proxy);
+        g_variant_unref (ret);
         return 1;
     }
 
-    dbus_g_connection_register_g_object (priv->connection,
-                                         "/org/gnome/Yelp",
-                                         G_OBJECT (app));
+    introspection_data = g_dbus_node_info_new_for_xml (introspection_xml, NULL);
+    g_dbus_connection_register_object (priv->connection,
+                                       "/org/gnome/Yelp",
+                                       introspection_data->interfaces[0],
+                                       &yelp_dbus_vtable,
+                                       app, NULL, NULL);
 
     application_setup (app);
 
@@ -380,6 +411,41 @@ yelp_application_run (YelpApplication  *app,
     return 0;
 }
 
+static void
+yelp_application_method (GDBusConnection *connection,
+                         const gchar *sender,
+                         const gchar *object_path,
+                         const gchar *interface_name,
+                         const gchar *method_name,
+                         GVariant *parameters,
+                         GDBusMethodInvocation *invocation,
+                         YelpApplication *app)
+{
+    if (g_str_equal (interface_name, "org.gnome.Yelp")) {
+        if (g_str_equal (method_name, "LoadUri")) {
+            GError *error = NULL;
+            gchar *uri;
+            guint32 timestamp;
+            g_variant_get (parameters, "(&su)", &uri, &timestamp);
+            yelp_application_load_uri (app, uri, timestamp, &error);
+            if (error) {
+                g_dbus_method_invocation_return_error (invocation,
+                                                       error->domain,
+                                                       error->code,
+                                                       "%s", error->message);
+                g_error_free (error);
+            }
+            else
+                g_dbus_method_invocation_return_value (invocation, NULL);
+            return;
+        }
+    }
+    g_dbus_method_invocation_return_error (invocation,
+                                           G_IO_ERROR,
+                                           G_IO_ERROR_NOT_SUPPORTED,
+                                           "Method not supported");
+}
+
 gboolean
 yelp_application_load_uri (YelpApplication  *app,
                            const gchar      *uri,
@@ -441,7 +507,7 @@ application_uri_resolved (YelpUri             *uri,
 
     doc_uri = yelp_uri_get_document_uri (uri);
 
-    if (data->new)
+    if (data->new || !doc_uri)
         window = NULL;
     else
         window = g_hash_table_lookup (priv->windows_by_document, doc_uri);
@@ -503,9 +569,9 @@ application_get_doc_settings (YelpApplication *app, const gchar *doc_uri)
         tmp = g_uri_escape_string (doc_uri, "", FALSE);
         settings_path = g_strconcat ("/apps/yelp/documents/", tmp, "/", NULL);
         g_free (tmp);
-        if (priv->gsettings_context)
-            settings = g_settings_new_with_context_and_path ("org.gnome.yelp.documents",
-                                                             priv->gsettings_context,
+        if (priv->backend)
+            settings = g_settings_new_with_backend_and_path ("org.gnome.yelp.documents",
+                                                             priv->backend,
                                                              settings_path);
         else
             settings = g_settings_new_with_path ("org.gnome.yelp.document",
@@ -518,8 +584,8 @@ application_get_doc_settings (YelpApplication *app, const gchar *doc_uri)
 
 static gboolean
 application_window_deleted (YelpWindow      *window,
-                              GdkEvent        *event,
-                              YelpApplication *app)
+                            GdkEvent        *event,
+                            YelpApplication *app)
 {
     gchar *doc_uri; /* owned by windows_by_document */
     YelpApplicationPrivate *priv = GET_PRIV (app);
@@ -548,36 +614,6 @@ application_maybe_quit (YelpApplication *app)
 
 /******************************************************************************/
 
-static void
-packages_installed (DBusGProxy     *proxy,
-                    DBusGProxyCall *call,
-                    gpointer        data)
-{
-    GError *error = NULL;
-    dbus_g_proxy_end_call (proxy, call, &error, G_TYPE_INVALID);
-    g_strfreev ((gchar **) data);
-    if (error) {
-        const gchar *err = NULL;
-        if (error->domain == DBUS_GERROR) {
-            if (error->code == DBUS_GERROR_SERVICE_UNKNOWN) {
-                err = _("You do not have PackageKit installed.  Package installation links require PackageKit.");
-            }
-            else if (error->code != DBUS_GERROR_REMOTE_EXCEPTION &&
-                     error->code != DBUS_GERROR_NO_REPLY) {
-                err = error->message;
-            }
-        }
-        if (err) {
-            GtkWidget *dialog = gtk_message_dialog_new (NULL, 0,
-                                                        GTK_MESSAGE_ERROR,
-                                                        GTK_BUTTONS_CLOSE,
-                                                        "%s", err);
-            gtk_dialog_run ((GtkDialog *) dialog);
-            gtk_widget_destroy (dialog);
-        }
-    }
-}
-
 void
 yelp_application_add_bookmark (YelpApplication   *app,
                                const gchar       *doc_uri,
@@ -616,6 +652,32 @@ yelp_application_add_bookmark (YelpApplication   *app,
     }
 }
 
+void
+yelp_application_remove_bookmark (YelpApplication   *app,
+                                  const gchar       *doc_uri,
+                                  const gchar       *page_id)
+{
+    GSettings *settings;
+
+    settings = application_get_doc_settings (app, doc_uri);
+
+    if (settings) {
+        GVariantBuilder builder;
+        GVariantIter *iter;
+        gchar *this_id, *this_icon, *this_title;
+        g_settings_get (settings, "bookmarks", "a(sss)", &iter);
+        g_variant_builder_init (&builder, G_VARIANT_TYPE ("a(sss)"));
+        while (g_variant_iter_loop (iter, "(&s&s&s)", &this_id, &this_icon, &this_title)) {
+            if (!g_str_equal (page_id, this_id))
+                g_variant_builder_add (&builder, "(sss)", this_id, this_icon, this_title);
+        }
+        g_variant_iter_free (iter);
+
+        g_settings_set_value (settings, "bookmarks", g_variant_builder_end (&builder));
+        g_signal_emit (app, signals[BOOKMARKS_CHANGED], 0, doc_uri);
+    }
+}
+
 void
 yelp_application_update_bookmarks (YelpApplication   *app,
                                    const gchar       *doc_uri,
@@ -669,26 +731,127 @@ yelp_application_get_bookmarks (YelpApplication *app,
     return g_settings_get_value (settings, "bookmarks");
 }
 
+static void
+packages_installed (GDBusConnection *connection,
+                    GAsyncResult *res,
+                    YelpApplication *app)
+{
+    GError *error = NULL;
+    g_dbus_connection_call_finish (connection, res, &error);
+    if (error) {
+        const gchar *err = NULL;
+        if (error->domain == G_DBUS_ERROR) {
+            if (error->code == G_DBUS_ERROR_SERVICE_UNKNOWN)
+                err = _("You do not have PackageKit installed.  Package installation links require PackageKit.");
+            else
+                err = error->message;
+        }
+        if (err != NULL) {
+            GtkWidget *dialog = gtk_message_dialog_new (NULL, 0,
+                                                        GTK_MESSAGE_ERROR,
+                                                        GTK_BUTTONS_CLOSE,
+                                                        "%s", err);
+            gtk_dialog_run ((GtkDialog *) dialog);
+            gtk_widget_destroy (dialog);
+        }
+        g_error_free (error);
+    }
+}
+
+void
+yelp_application_add_read_later (YelpApplication   *app,
+                                 const gchar       *doc_uri,
+                                 const gchar       *full_uri,
+                                 const gchar       *title)
+{
+    GSettings *settings;
+
+    settings = application_get_doc_settings (app, doc_uri);
+
+    if (settings) {
+        GVariantBuilder builder;
+        GVariantIter *iter;
+        gchar *this_uri, *this_title;
+        gboolean broken = FALSE;
+        g_settings_get (settings, "readlater", "a(ss)", &iter);
+        g_variant_builder_init (&builder, G_VARIANT_TYPE ("a(ss)"));
+        while (g_variant_iter_loop (iter, "(&s&s)", &this_uri, &this_title)) {
+            if (g_str_equal (full_uri, this_uri)) {
+                /* Already have this link */
+                broken = TRUE;
+                break;
+            }
+            g_variant_builder_add (&builder, "(ss)", this_uri, this_title);
+        }
+        g_variant_iter_free (iter);
+
+        if (!broken) {
+            GVariant *value;
+            g_variant_builder_add (&builder, "(ss)", full_uri, title);
+            value = g_variant_builder_end (&builder);
+            g_settings_set_value (settings, "readlater", value);
+            g_signal_emit (app, signals[READ_LATER_CHANGED], 0, doc_uri);
+        }
+    }
+}
+
+void
+yelp_application_remove_read_later (YelpApplication *app,
+                                    const gchar     *doc_uri,
+                                    const gchar     *full_uri)
+{
+    GSettings *settings;
+
+    settings = application_get_doc_settings (app, doc_uri);
+
+    if (settings) {
+        GVariantBuilder builder;
+        GVariantIter *iter;
+        gchar *this_uri, *this_title;
+        g_settings_get (settings, "readlater", "a(ss)", &iter);
+        g_variant_builder_init (&builder, G_VARIANT_TYPE ("a(ss)"));
+        while (g_variant_iter_loop (iter, "(&s&s)", &this_uri, &this_title)) {
+            if (!g_str_equal (this_uri, full_uri))
+                g_variant_builder_add (&builder, "(ss)", this_uri, this_title);
+        }
+        g_variant_iter_free (iter);
+
+        g_settings_set_value (settings, "readlater", g_variant_builder_end (&builder));
+        g_signal_emit (app, signals[READ_LATER_CHANGED], 0, doc_uri);
+    }
+}
+
+GVariant *
+yelp_application_get_read_later (YelpApplication *app,
+                                 const gchar     *doc_uri)
+{
+    GSettings *settings = application_get_doc_settings (app, doc_uri);
+
+    return g_settings_get_value (settings, "readlater");
+}
+
 void
 yelp_application_install_package (YelpApplication  *app,
                                   const gchar      *pkg,
                                   const gchar      *alt)
 {
+    GVariantBuilder *strv;
     YelpApplicationPrivate *priv = GET_PRIV (app);
     guint32 xid = 0;
-    DBusGProxy *proxy = dbus_g_proxy_new_for_name (priv->connection,
-                                                   "org.freedesktop.PackageKit",
-                                                   "/org/freedesktop/PackageKit",
-                                                   "org.freedesktop.PackageKit.Modify");
-    gchar **pkgs = g_new0 (gchar *, 2);
-    pkgs[0] = g_strdup (pkg);
-
-    dbus_g_proxy_begin_call (proxy, "InstallPackageNames",
-                             packages_installed, pkgs, NULL,
-                             G_TYPE_UINT, xid,
-                             G_TYPE_STRV, pkgs,
-                             G_TYPE_STRING, "",
-                             G_TYPE_INVALID, G_TYPE_INVALID);
+    strv = g_variant_builder_new (G_VARIANT_TYPE ("as"));
+    g_variant_builder_add (strv, "s", pkg);
+    g_dbus_connection_call (priv->connection,
+                            "org.freedesktop.PackageKit",
+                            "/org/freedesktop/PackageKit",
+                            "org.freedesktop.PackageKit.Modify",
+                            "InstallPackageNames",
+                            g_variant_new ("(uass)", xid, strv, ""),
+                            NULL,
+                            G_DBUS_CALL_FLAGS_NONE,
+                            -1, NULL,
+                            (GAsyncReadyCallback) packages_installed,
+                            app);
+    g_variant_builder_unref (strv);
 }
 
 static gboolean
